
/* TODO

   - PBM output
*/

/* HP-GL Interpreter Version 3.5
   Reads HP-GL (usually *.plt) files and draws the graphics on screen. */

#include <stdio.h>
#include <graphics.h>
#include <conio.h>
#include <string.h>
#include <ctype.h>
#include "input2.h"
#include "box.h"
#include "mytypes.h"
#include "buffered_io.h"
#include "kernel.h"
#include "pbmout.h"

#define MODE_SET   0
#define PAUSE_MODE true
#define START      'S'
#define MNEMONIC   'M'
#define PARAMETER  'P'
#define ERROR      'E'
#define UNKNOWN    0
#define INIT       1
#define PEN_DOWN   2
#define PEN_UP     3
#define POS_ABS    4
#define POS_REL    5
#define MIN_COORD -10000
#define MAX_COORD  10000

char ch;
char state;
int i;
char fileName[15];
char fileNameCfg[15];
char fileNamePBM[15];
int command;
char firstMnemCh;
int value;
bool isNegative;
bool isPenDown;
int lastX;
int lastY;
int paramX;
int paramY;
char paramName;
int dotCnt;
int maxX, maxY;
int minX, minY;
int offsetX;
int offsetY;
float scaleX;
float scaleY;
int pass;
uint8_t row;
bool createPbm;
char arCorr;

// forward declarations
// void printCommand();  // TEST
void scaleParameter();
void drawLine();


/* Initializes a parameter, i.e. resets all necessary variables
   before a new parameter can be read in. */
void initParameter() {
  value = 0;
  isNegative = false;
}


/* Checks if a character is a whitespace. */
bool isWhitespace(char c) {
  return c < 33;
}


/* Finishes a parameter, i.e. all the work that has to be done after
   an HP-GL parameter has been read in completely. */
void finishParameter() {
  if (command > INIT && command < 6) {
    if (isNegative) value = -value;
    if (paramName == 'X') {
      lastX = paramX;
      paramX = value;
      if (pass == 1) {
        if (value > maxX) maxX = value;
        if (value < minX) minX = value;
      }
      paramName = 'Y';
    }
    else {
      lastY = paramY;
      paramY = value;
      scaleParameter();
      if (pass == 1) {
        if (value > maxY) maxY = value;
        if (value < minY) minY = value;
      }
      paramName = 'X';
      if (pass == 2 && isPenDown) drawLine();
    }
  }
  initParameter();
}


/* Scales the X and Y HP-GL parameters, transforming them to screen
   coordinates. */
void scaleParameter() {
  float r;

  r = paramX - minX;
  r = r * scaleX;
  paramX = (int) r + offsetX;

  r = maxY - paramY;
  r = r * scaleY;
  paramY = (int) r + offsetY;
}


/* Prints a dot every 10 rounds. */
void printDot() {
  if (pass == 1) {
    --dotCnt;
    if (dotCnt == 0) {
      dotCnt = 10;
      putchar('.');
    }
  }
}


/* Processes a character depending on the current state of the state machine. */
void processChar() {

  if (ch == ';') {
    state = START;
    finishParameter();
    paramName = 'X';
  }
  else {
    switch (state) {
      case START:
          printDot();
          command = UNKNOWN;
          initParameter();
          if (!isWhitespace(ch)) {
            if (ch >= 'A' && ch <= 'Z') {
              firstMnemCh = ch;
              state = MNEMONIC;
            }
            else {
              printf("\nInvalid first character in Mnemonic: %c\n", ch);
              state = ERROR;
            }
          }
          break;
      case MNEMONIC:
          if (ch >= 'A' && ch <= 'Z') {
            switch (firstMnemCh) {
              case 'I':
                   switch (ch) {
                     case 'N': command = INIT;     isPenDown = false; break;
                   }
              case 'P':
                   switch(ch) {
                     case 'D': command = PEN_DOWN; isPenDown = true;  break;
                     case 'U': command = PEN_UP;   isPenDown = false; break;
                     case 'A': command = POS_ABS;                     break;
                     case 'R': command = POS_REL;                     break;
                   }
            }
            // printCommand(); // Test
            state = PARAMETER;
          }
          else {
            printf("\nInvalid second character in Mnemonic: %c\n", ch);
            state = ERROR;
          }
          break;
      case PARAMETER:
          if (ch != '.') { // real values not supported; dot ignored
            if (ch == '-') {
              isNegative = true;
            }
            else if (ch >= '0' && ch <= '9') {
              value = value * 10 + ch - '0';
            }
            else {
              // delimiter found: end parameter
              finishParameter();
            }
          }
          break;
    }
  }
}


/* TEST >>>
void printCommand() {
  switch (command) {
    case UNKNOWN:  printf("UNKNOWN");
    case PEN_DOWN: printf("PEN_DOWN");
    case PEN_UP:   printf("PEN_UP");
    case POS_ABS:  printf("POS_ABS");
    case POS_REL:  printf("POS_REL");
  }
} <<< TEST */


/* Draws a line; includes a parameter check to prevent overflow. */
void drawLine() {

  if (lastX >= 0 && lastX < 720 &&
      lastY >= 0 && lastY < 256 &&
      paramX >= 0 && paramX < 720 && 
      paramY >= 0 && paramY < 256) {
    draw(lastX, lastY, paramX, paramY);
  }
}


/* Calculates the scaling of the picture with the correct aspect ratio,
   taking into account the correct pixel aspect ratio. */
void calcScaling() {
  int height, width;
  float h2, w2;

  width  = maxX - minX;
  height = maxY - minY;
  if (height > width) {
    scaleX = 540.0 / (float) height;
    scaleY = scaleX * 256.0 / 540.0;
    w2 = width * scaleX;
    offsetX = (720 - (int) w2) / 2;
    offsetY = 0;
  }
  else {
    scaleX = 720.0 / (float) width;
    scaleY = scaleX * 256.0 / 540.0;
    h2 = height * scaleY;
    offsetX = 0;
    offsetY = (256 - (int) h2) / 2;
  }
}


/* Creates a filename for the configuration file. It has the same name as
   the HP-GL file, but width the extension ".cfg". */
void createConfigFileName() {
  int i;
  char ch;
  int len;

  i = 0;
  len = strlen(fileName);

  do { // copy filename before the dot
    ch = fileName[i];
    fileNameCfg[i] = ch;
    ++i;
  } while (i < len && ch != '.');
  
  fileNameCfg[i] = '\0'; // finalize string

  if (ch == '.') { // add cfg extension
    strcat(fileNameCfg, "cfg");
  }
}


/* First pass: acquire mininum and maximum parameter values. */
void pass1() {
  FILE *filVarCfg;

  pass = 1;
  
  puts("First pass: determining min. and max. X and Y coordinates.");
  
  createConfigFileName();

  filVarCfg = fopen(fileNameCfg, "r");
  if (filVarCfg == NULL) {
    printf("Configuration file not found: %s\n", fileNameCfg);
    writeKeyMessage();
    waitForKey();
    // open cfg file for write access
    fclose(filVarCfg);
    filVarCfg = fopen(fileNameCfg, "w");
	
	// get min and max parameter values
    for (;;) {
      ch = (char) readBufferedByte();
      if (endOfBufferedFile()) {
          break;
      }
      processChar();
    }

	// write min and max values to configuration file
    fprintf(filVarCfg, "%d\n", minX);
    fprintf(filVarCfg, "%d\n", maxX);
    fprintf(filVarCfg, "%d\n", minY);
    fprintf(filVarCfg, "%d\n", maxY);
  }
  else {
    // read cfg file
    fscanf(filVarCfg, "%d\n", &minX);
    fscanf(filVarCfg, "%d\n", &maxX);
    fscanf(filVarCfg, "%d\n", &minY);
    fscanf(filVarCfg, "%d\n", &maxY);
    puts("\nConfiguration read:");
    printf("Min. X: %6d   Max. X: %6d\n", minX, maxX);
    printf("Min. Y: %6d   Max. Y: %6d\n", minY, maxY);
    writeKeyMessage();
    waitForKey();
  }

  fclose(filVarCfg);
}


/* Second pass: draw the picture. */
void pass2() {

  errorCode_t errCode;
  
  pass = 2;
  
  clearScreen();
  puts("Second pass: drawing\n");
  printf("Min. X: %6d   Max. X: %6d\n", minX, maxX);
  printf("Min. Y: %6d   Max. Y: %6d\n", minY, maxY);
  putchar('\n');
  calcScaling();
  printf("Scale  X: %6.3f   Scale  Y: %6.3f\n", scaleX, scaleY);
  printf("Offset X: %6d   Offset Y: %6d\n", offsetX, offsetY);
  writeKeyMessage();
  waitForKey();

  closeBufferedFile();
  errCode = openBufferedFile(fileName, "rb");
  
  clearScreen();

  for (;;) {
    ch = (char) readBufferedByte();
    if (endOfBufferedFile()) {
        break;
    }
    processChar();
  }
}


/* Main program */
int main(int argc, char *argv[]) {

  errorCode_t errCode;
  
  statusBarOff();
  clearScreen();
  gx_init();
  
  drawBox(0, 0, 90, 3); gotoXY(26, 1);
  printf("HP-GL interpreter V3.5 by Bernd Bock");
  createPbm = false;
  
  // get the name of the HP-GL file
  if (argc > 1) {
    strcpy(fileName, argv[1]); // first commandline parameter treated as filename
    gotoXY(0, 4);
  }
  else {
    drawBox( 0, 3, 52, 3);
    drawBox(52, 3, 38, 3);
    drawBox( 0, 6, 52, 3);
    drawBox(52, 6, 38, 3);
    gotoXY(2, 7);
    printf("Enter PBM output filename:");
    gotoXY(54, 7);
    printf("AR correction? (Y/N)");
    gotoXY(2, 4);
    printf("Enter name of HPGL file to draw: ");
    inputString(fileName, 14, 35, 4, FTFileName);
    if (*fileName == '\0') goto byebye; // abort if no filename entered
    //gotoXY(35, 7);
    inputString(fileNamePBM, 14, 35, 7, FTFileName);
    if (*fileNamePBM != '\0') {
      createPbm = true;
      gotoXY(75, 7);
      do {
        arCorr = toupper(getKey());
      } while (arCorr != 'Y' && arCorr != 'N');
      gotoXY(75, 7);
      putchar(arCorr);
    }
    gotoXY(0, 9);
  }
  // open the HP-GL file
  errCode = openBufferedFile(fileName, "rb");
  if (errCode == ErrCouldNotOpenFile) {
    printf("File not found: %s\n", fileName);
    goto byebye;
  }

  // initialization
  state = START;
  paramName = 'X';
  lastX = 0;
  lastY = 0;
  minX = MAX_COORD;
  maxX = MIN_COORD;
  minY = MAX_COORD;
  maxY = MIN_COORD;
  dotCnt = 10;

  hideCursor();
  putchar('\n');;

  pass1();
  pass2();

  closeBufferedFile();

  // save picture as PBM file
  if (createPbm) {
      saveScreenToPBM(fileNamePBM, arCorr == 'Y');
  }
  
  beep();
  waitForKey();
  
byebye:
  statusBarOn();
  showCursor();
}
